@@ -1,7 +1,9 @@
-use cssparser::{CowRcStr, Parser, Token};
+//! `userSpaceOnUse` or `objectBoundingBox` values.
 
-use crate::error::ValueErrorKind;
-use crate::parsers::{Parse, ParseError};
+use cssparser::Parser;
+
+use crate::error::*;
+use crate::parsers::Parse;
 
 /// Defines the units to be used for things that can consider a
 /// coordinate system in terms of the current transformation, or in
@@ -13,27 +15,12 @@ pub enum CoordUnits {
 }
 
 impl Parse for CoordUnits {
-    type Err = ValueErrorKind;
-
-    fn parse(parser: &mut Parser<'_, '_>) -> Result<CoordUnits, ValueErrorKind> {
-        let loc = parser.current_source_location();
-
-        parser
-            .expect_ident()
-            .and_then(|cow| match cow.as_ref() {
-                "userSpaceOnUse" => Ok(CoordUnits::UserSpaceOnUse),
-                "objectBoundingBox" => Ok(CoordUnits::ObjectBoundingBox),
-                _ => Err(
-                    loc.new_basic_unexpected_token_error(Token::Ident(CowRcStr::from(
-                        cow.as_ref().to_string(),
-                    ))),
-                ),
-            })
-            .map_err(|_| {
-                ValueErrorKind::Parse(ParseError::new(
-                    "expected 'userSpaceOnUse' or 'objectBoundingBox'",
-                ))
-            })
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Self, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "userSpaceOnUse" => CoordUnits::UserSpaceOnUse,
+            "objectBoundingBox" => CoordUnits::ObjectBoundingBox,
+        )?)
     }
 }
 
@@ -64,11 +51,9 @@ macro_rules! coord_units {
         }
 
         impl $crate::parsers::Parse for $name {
-            type Err = $crate::error::ValueErrorKind;
-
-            fn parse(
-                parser: &mut ::cssparser::Parser<'_, '_>,
-            ) -> Result<Self, $crate::error::ValueErrorKind> {
+            fn parse<'i>(
+                parser: &mut ::cssparser::Parser<'i, '_>,
+            ) -> Result<Self, $crate::error::ParseError<'i>> {
                 Ok($name($crate::coord_units::CoordUnits::parse(parser)?))
             }
         }
